from lightwood.encoders.categorical.onehot import OneHotEncoder
from lightwood.encoders.categorical.autoencoder import CategoricalAutoEncoder

default = CategoricalAutoEncoder
onehot = OneHotEncoder
autoencoder = CategoricalAutoEncoder
